use std::fmt;
use std::io::Read;
use std::net;
use std::net::SocketAddr;
use std::pin::Pin;

use ansi_term::{Color, Style};
use error::*;
use futures_util::{future::TryFutureExt, stream::Stream};
use hyper::server::conn::AddrStream;
use hyper::service::{make_service_fn, service_fn};
use hyper::Server;
use rustls_pemfile;

mod database;
mod error;
mod http;
mod usimp;
mod websocket;

struct HyperAcceptor<'a> {
    acceptor: Pin<
        Box<
            dyn Stream<
                    Item = Result<
                        tokio_rustls::server::TlsStream<tokio::net::TcpStream>,
                        std::io::Error,
                    >,
                > + 'a,
        >,
    >,
}

impl hyper::server::accept::Accept for HyperAcceptor<'_> {
    type Conn = tokio_rustls::server::TlsStream<tokio::net::TcpStream>;
    type Error = std::io::Error;

    fn poll_accept(
        mut self: Pin<&mut Self>,
        cx: &mut core::task::Context,
    ) -> core::task::Poll<Option<Result<Self::Conn, Self::Error>>> {
        Pin::new(&mut self.acceptor).poll_next(cx)
    }
}

fn load_certs(filename: &str) -> std::io::Result<Vec<rustls::Certificate>> {
    let certfile = std::fs::File::open(filename)
        .map_err(|e| error(format!("failed to open {}: {}", filename, e)))?;
    let mut reader = std::io::BufReader::new(certfile);

    Ok(rustls_pemfile::certs(&mut reader)
        .unwrap()
        .iter()
        .map(|v| rustls::Certificate(v.clone()))
        .collect())
}

fn load_private_key(filename: &str) -> std::io::Result<rustls::PrivateKey> {
    let keyfile = std::fs::File::open(filename)
        .map_err(|e| error(format!("failed to open {}: {}", filename, e)))?;
    let mut reader = std::io::BufReader::new(keyfile);

    loop {
        match rustls_pemfile::read_one(&mut reader).expect("cannot parse private key .pem file") {
            Some(rustls_pemfile::Item::RSAKey(key)) => return Ok(rustls::PrivateKey(key)),
            Some(rustls_pemfile::Item::PKCS8Key(key)) => return Ok(rustls::PrivateKey(key)),
            Some(rustls_pemfile::Item::ECKey(key)) => return Ok(rustls::PrivateKey(key)),
            None => break,
            _ => {}
        }
    }

    Err(error("unexpected error".into()))
}

fn error(err: String) -> std::io::Error {
    std::io::Error::new(std::io::ErrorKind::Other, err)
}

#[tokio::main]
async fn main() -> Result<(), Error> {
    println!("Locutus server");

    database::init().await?;
    usimp::subscription::init();

    let server1 = Server::bind(&SocketAddr::from((
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        8080,
    )));
    let service = make_service_fn(|_: &AddrStream| async {
        Ok::<_, hyper::Error>(service_fn(http::handler))
    });
    let srv1 = server1.serve(service);

    let tls_cfg = {
        let certs = load_certs("/home/lorenz/Certificates/priv/necronda.net/fullchain.pem").unwrap();
        let key = load_private_key("/home/lorenz/Certificates/priv/necronda.net/privkey.pem").unwrap();
        let mut cfg = rustls::ServerConfig::builder()
            .with_safe_default_cipher_suites()
            .with_safe_default_kx_groups()
            .with_protocol_versions(rustls::ALL_VERSIONS)
            .unwrap()
            .with_no_client_auth()
            .with_single_cert(certs, key)
            .unwrap();
        cfg.alpn_protocols = vec![b"h2".to_vec(), b"http/1.1".to_vec()];
        std::sync::Arc::new(cfg)
    };

    let acceptor = tokio_rustls::TlsAcceptor::from(tls_cfg);
    let tcp = Box::pin(tokio::net::TcpListener::bind("[::]:8443").await.unwrap());
    let incoming_tls_stream = async_stream::stream! {
        loop {
            let (socket, _) = tcp.accept().await.unwrap();
            let stream = acceptor.accept(socket).map_err(|e| {
                println!("[!] Voluntary server halt due to client-connection error...");
                // Errors could be handled here, instead of server aborting.
                //Ok(None)
                //println!("{:?}", e);
                error(format!("TLS Error: {:?}", e))
            });
            yield stream.await;
        }
    };

    let server2 = Server::builder(HyperAcceptor {
        acceptor: Box::pin(incoming_tls_stream),
    });

    let service = make_service_fn(|_| async { Ok::<_, hyper::Error>(service_fn(http::handler)) });
    let srv2 = server2.serve(service);

    println!("{}", Color::Green.paint("Ready"));

    let (_res1, _res2) = futures::future::join(srv1, srv2).await;

    Ok(())
}
