use crate::error::*;
use crate::usimp;
use crate::usimp::*;
use futures::stream::{SplitSink, SplitStream};
use futures_util::SinkExt;
use futures_util::StreamExt;
use hyper::{header, Body, Request, StatusCode};
use hyper_tungstenite::hyper::upgrade::Upgraded;
use hyper_tungstenite::tungstenite::{handshake, Message};
use hyper_tungstenite::{tungstenite::protocol::Role, WebSocketStream};
use serde_json::{Map, Value};
use tokio::sync::mpsc;

pub enum WebSocketEnvelope {
    Close,
    Ping(Vec<u8>),
    Pong(Vec<u8>),
    Text(OutputEnvelope),
}

impl From<OutputEnvelope> for WebSocketEnvelope {
    fn from(envelope: OutputEnvelope) -> Self {
        WebSocketEnvelope::Text(envelope)
    }
}

async fn sender(
    mut sink: SplitSink<WebSocketStream<Upgraded>, Message>,
    mut rx: mpsc::Receiver<WebSocketEnvelope>,
) {
    while let Some(msg) = rx.recv().await {
        match msg {
            WebSocketEnvelope::Ping(data) => {
                if let Err(error) = sink.send(Message::Ping(data)).await {
                    eprintln!("{:?}", error);
                    return;
                }
            }
            WebSocketEnvelope::Pong(data) => {
                if let Err(error) = sink.send(Message::Pong(data)).await {
                    eprintln!("{:?}", error);
                    return;
                }
            }
            WebSocketEnvelope::Close => {
                if let Err(error) = sink.send(Message::Close(None)).await {
                    eprintln!("{:?}", error);
                    return;
                }
                break;
            }
            WebSocketEnvelope::Text(msg) => {
                let mut envelope = Value::Object(Map::new());
                envelope["data"] = msg.data;
                envelope["request_nr"] = match msg.request_nr {
                    Some(nr) => Value::from(nr),
                    None => Value::Null,
                };
                match msg.error {
                    Some(error) => {
                        envelope["status"] = Value::from("error");
                        envelope["error"] = Value::from(error);
                    }
                    None => {
                        envelope["status"] = Value::from("success");
                    }
                }
                if let Err(error) = sink.send(Message::Text(envelope.to_string())).await {
                    eprintln!("{:?}", error);
                    return;
                }
            }
        }
    }
}

async fn receiver(
    mut stream: SplitStream<WebSocketStream<Upgraded>>,
    tx: mpsc::Sender<WebSocketEnvelope>,
) {
    while let Some(res) = stream.next().await {
        match res {
            Ok(msg) => {
                let _res;
                if msg.is_ping() {
                    _res = tx.send(WebSocketEnvelope::Pong(msg.into_data())).await;
                } else if msg.is_pong() {
                    // Ignore
                } else if msg.is_close() {
                    _res = tx.send(WebSocketEnvelope::Close).await;
                    break;
                } else if msg.is_binary() {
                    _res = tx.send(WebSocketEnvelope::Text(OutputEnvelope {
                        error: Some(Error {
                            kind: ErrorKind::WebSocketError,
                            class: ErrorClass::ClientProtocolError,
                            msg: Some("Binary frames are not allowed".to_string()),
                            desc: None,
                        }),
                        request_nr: None,
                        data: Value::Null,
                    })).await;
                } else if msg.is_text() {
                    let input: InputEnvelope = serde_json::from_slice(&msg.into_data()[..]).unwrap();
                    let output = match usimp::endpoint(&input, Some(tx.clone())).await {
                        Ok(output) => output,
                        Err(error) => input.error(error),
                    };
                    _res = tx.send(WebSocketEnvelope::Text(output)).await;
                } else {
                    _res = tx.send(WebSocketEnvelope::Text(OutputEnvelope {
                        error: Some(Error {
                            kind: ErrorKind::WebSocketError,
                            class: ErrorClass::ClientProtocolError,
                            msg: Some("Unknown frame opcode".to_string()),
                            desc: None,
                        }),
                        request_nr: None,
                        data: Value::Null,
                    })).await;
                }
            }
            Err(error) => println!("{:?}", error),
        }
    }
}

pub async fn handler(
    req: Request<Body>,
    res: hyper::http::response::Builder,
) -> (hyper::http::response::Builder, Option<OutputEnvelope>) {
    match req.headers().get(header::UPGRADE) {
        Some(val) if val == header::HeaderValue::from_str("websocket").unwrap() => {}
        _ => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::WebSocketError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            );
        }
    }

    let key = match req.headers().get(header::SEC_WEBSOCKET_KEY) {
        Some(key) => key,
        None => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::WebSocketError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            );
        }
    };
    let key = handshake::derive_accept_key(key.as_bytes());

    match req.headers().get(header::SEC_WEBSOCKET_PROTOCOL) {
        Some(val) if val == header::HeaderValue::from_str("usimp").unwrap() => {}
        _ => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::UsimpError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            );
        }
    }

    tokio::spawn(async move {
        match hyper::upgrade::on(req).await {
            Ok(upgraded) => {
                let ws_stream = WebSocketStream::from_raw_socket(upgraded, Role::Server, None).await;
                let (tx, rx) = mpsc::channel::<WebSocketEnvelope>(64);
                let (sink, stream) = ws_stream.split();

                let other = tokio::spawn(async move { sender(sink, rx).await });
                receiver(stream, tx).await;
                other.await.unwrap();
            }
            Err(error) => eprintln!("Unable to upgrade: {}", error),
        }
    });

    (
        res.status(StatusCode::SWITCHING_PROTOCOLS)
            .header(
                header::CONNECTION,
                header::HeaderValue::from_str("Upgrade").unwrap(),
            )
            .header(
                header::UPGRADE,
                header::HeaderValue::from_str("websocket").unwrap(),
            )
            .header(
                header::SEC_WEBSOCKET_ACCEPT,
                header::HeaderValue::from_str(key.as_str()).unwrap(),
            )
            .header(
                header::SEC_WEBSOCKET_PROTOCOL,
                header::HeaderValue::from_str("usimp").unwrap(),
            ),
        None,
    )
}
